# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Wang Xiao <xiawang3@cisco.com>, Rudy Lei <shlei@cisco.com>

import os
import re
import sys
import click
import shutil
import errorhandler

from .. import __version__
from . import options, selections, cli_validators
from aac_init.conf import settings
from aac_init.log_utils import setup_logger


error_handler = errorhandler.ErrorHandler()


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(__version__)
@options.yaml_dir_path
@options.log_level
def main(data: str, log_level: str) -> None:
    """
    A CLI tool to bootstrap and configure ACI fabric using ACI as Code.
    """
    # Setup working dir
    output_path = settings.OUTPUT_BASE_DIR

    if os.path.exists(output_path) and os.path.isdir(output_path):
        shutil.rmtree(output_path)

    settings.DEFAULT_LOG_LEVEL = log_level

    logger = setup_logger("main.log")

    cli_validator = cli_validators.CLI_Validator(data, output_path)
    logger.debug("CLI Validator initialized.\n")

    # Type single number or multiple numbers (1,2... or *)
    selection_prompt = (
        "Select single or multiple choice(s) "
        "to init ACI Fabric:\n{}\nExample: (1,2,.. or *)".format(
            "\n".join(
                [
                    f"[{i + 1}]  {selection}"
                    for i, selection in enumerate(settings.DEFAULT_USER_SELECTIONS)
                ]
            )
        )
    )
    selection_choices = click.prompt(
        click.style(selection_prompt, fg="green"),
        type=cli_validator.validate_selections,
    )
    logger.debug("Selections validated successfully.")
    if not selection_choices:
        logger.debug("Selections validated failed. Exiting...")
        exit()

    # Type "yes" or "no" to confirm
    confirm_prompt = (
        "\nAre you sure to proceed with the following choice(s)?\n"
        + "\n".join(
            f"[{i}] {settings.DEFAULT_USER_SELECTIONS[int(i) - 1]}"
            for i in selection_choices
        )
        + "\n"
    )

    selections_confirm = click.prompt(
        click.style(confirm_prompt, fg="green"),
        type=click.Choice(["yes", "no"], case_sensitive=False),
        default="yes",
        show_default=True,
        show_choices=True,
    )

    if not re.match(r"yes", selections_confirm, re.IGNORECASE):
        logger.error("Process aborted, exiting aac-init tool...")
        exit()

    logger.info("Start to process aac-init tool!")
    logger.info("Validating data file...")
    if cli_validator.validate_cli_input():
        logger.info("Input data files validated successfully.")
    else:
        logger.error("input data files validated failed!")
        logger.error("Error details:\n")
        formatted_errors = "\n".join(
            [f"{i + 1}. {error}" for i, error in enumerate(cli_validator.errors)]
        )
        logger.error(f"Errors:\n{formatted_errors}")
        exit()

    # selections handling
    cli_selection = selections.Selections(data, output_path)
    for selection in selection_choices:
        logger.info(f"Start to process selection: {selection}")
        match selection:
            case "1":
                if cli_selection.fabric_bootstrap():
                    logger.info("ACI fabric bootstrap successfully.")
                else:
                    logger.error("ACI fabric bootstrap failed!")
                    exit()
            case "2":
                if cli_selection.apic_init_setup():
                    logger.info("APIC initial setup successfully.")
                else:
                    logger.error("APIC initial setup failed!")
                    exit()
            case "3":
                if cli_selection.apic_nac_config():
                    logger.info("Configure ACI fabric successfully.")
                else:
                    logger.error("Configure ACI fabric failed!")
                    exit()
            case _:
                logger.error(f"Unknown selection: {selection}")
                exit()


def exit() -> None:
    if error_handler.fired:
        sys.exit(1)
    else:
        sys.exit(0)
